@@ -24,6 +24,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <assert.h>
 
 #ifdef WINDOWS
 #include <WinFunctions.h>
@@ -42,6 +43,71 @@
 #define	WARNING_FILE_TOO_SMALL	2
 #define	WARNING_CANNOT_HASH	3
 
+#define CONVERT_TAB "_<tlsh_convert_tab>_"
+#define CONVERT_NEWLINE "_<tlsh_convert_newline>_"
+#define CONVERT_LINEFEED "_<tlsh_convert_linefeed>_"
+
+typedef enum {
+	TAB,
+	NEWLINE,
+	LINEFEED
+} SpecialChar;
+
+SpecialChar getSpecialChar(const char *tab, const char *newline, const char *linefeed)
+{
+    // To call this function, there has to be at least 1 special character
+	assert (tab != NULL || newline != NULL || linefeed != NULL);  
+
+	if (tab == NULL) return (newline == NULL) ? LINEFEED : (linefeed == NULL) ? NEWLINE : (newline < linefeed) ? NEWLINE : LINEFEED;
+	if (newline == NULL) return (tab == NULL) ? LINEFEED : (linefeed == NULL) ? TAB : (tab < linefeed) ? TAB : LINEFEED;
+	if (linefeed == NULL) return (tab == NULL) ? NEWLINE : (newline == NULL) ? TAB : (tab < newline) ? TAB : NEWLINE;
+}
+
+static const char *convert_special_chars(char *filename, char *buf, size_t bufSize, 
+                                         const char *curTab, const char *replaceTab,
+                                         const char *curNewline, const char *replaceNewline,
+                                         const char *curLinefeed, const char *replaceLinefeed)
+{
+	size_t fname_offset = 0;
+	size_t buf_offset = 0;
+
+	while (true) {
+		char *tab = strstr(filename+fname_offset, curTab);
+		char *newline = strstr(filename+fname_offset, curNewline);
+		char *linefeed = strstr(filename+fname_offset, curLinefeed);
+
+		// when no more special characters to replace, copy the remaining part of filename and then return.
+		if (tab == NULL && newline == NULL && linefeed == NULL) {
+			snprintf(buf+buf_offset, bufSize-buf_offset, "%s", filename+fname_offset);
+			return buf;
+		}
+
+		SpecialChar sp = getSpecialChar(tab, newline, linefeed);
+		if (sp == TAB) {
+			char save = *tab;
+			*tab = '\0';  // terminate for snprintf
+			buf_offset += snprintf(buf+buf_offset, bufSize-buf_offset, "%s%s", filename+fname_offset, replaceTab);
+			fname_offset = tab - filename + strlen(curTab);
+			*tab = save;  // replace tab
+		}
+		else if (sp == NEWLINE) {
+			char save = *newline;
+			*newline = '\0';  // terminate for snprintf
+			buf_offset += snprintf(buf+buf_offset, bufSize-buf_offset, "%s%s", filename+fname_offset, replaceNewline);
+			fname_offset = newline - filename + strlen(curNewline);
+			*newline = save;  // replace newline
+		}
+		else {
+			assert (sp == LINEFEED);
+			char save = *linefeed;
+			*linefeed = '\0';  // terminate for snprintf
+			buf_offset += snprintf(buf+buf_offset, bufSize-buf_offset, "%s%s", filename+fname_offset, replaceLinefeed);
+			fname_offset = linefeed - filename + strlen(curLinefeed);
+			*linefeed = save;  // replace linefeed
+		}
+	}
+}
+
 static int read_file_eval_tlsh(char *fname, Tlsh *th, int show_details)
 {
 	///////////////////////////////////////
@@ -58,7 +124,7 @@ static int read_file_eval_tlsh(char *fname, Tlsh *th, int show_details)
 
 	fclose(fd);
 
-	if (sizefile < 512)
+	if (sizefile < MIN_DATA_LENGTH)
 		return(WARNING_FILE_TOO_SMALL);
 
 	///////////////////////////////////////
@@ -146,10 +212,8 @@ struct dirent   *dit;
 	int n_file = 0;
 	while (dit != NULL) {
 		char tmp_fname[2000];
-		strncpy(tmp_fname,	dirname, sizeof(tmp_fname)-1);
-		strncat(tmp_fname,	"/", sizeof(tmp_fname)-1);
-		strncat(tmp_fname,	dit->d_name, sizeof(tmp_fname)-1);
-		if (strlen(tmp_fname) < sizeof(tmp_fname) - 2) {
+		int len = snprintf(tmp_fname,	sizeof(tmp_fname)-1, "%s/%s", dirname, dit->d_name);
+		if (len < sizeof(tmp_fname) - 2) {
 			if (is_dir(tmp_fname) ) {
 				if ((strcmp(dit->d_name, ".") == 0) || (strcmp(dit->d_name, "..") == 0)) {
 					;
@@ -180,11 +244,9 @@ struct dirent   *dit;
 	dit = readdir(dip);
 	while (dit != NULL) {
 		char tmp_fname[2000];
-		strncpy(tmp_fname,	dirname, sizeof(tmp_fname)-1);
-		strncat(tmp_fname,	"/", sizeof(tmp_fname)-1);
-		strncat(tmp_fname,	dit->d_name, sizeof(tmp_fname)-1);
+		int len = snprintf(tmp_fname,	sizeof(tmp_fname)-1, "%s/%s", dirname, dit->d_name);
 		// -2 for safety
-		if (strlen(tmp_fname) < sizeof(tmp_fname) - 2) {
+		if (len < sizeof(tmp_fname) - 2) {
 			if (is_dir(tmp_fname) ) {
 				if ((strcmp(dit->d_name, ".") == 0) || (strcmp(dit->d_name, "..") == 0)) {
 					;
@@ -323,7 +385,7 @@ int max_files;
 			}
 
 			fnames[count].tlsh = strdup(buf);
-			fnames[count].name = strdup(x);
+			fnames[count].name = strdup(convert_special_chars(x, buf, sizeof(buf), CONVERT_TAB, "\t", CONVERT_NEWLINE, "\n", CONVERT_LINEFEED, "\r"));
 
 			count ++;
 
@@ -415,6 +477,8 @@ int max_files;
 		}
 	}
 
+	char buf1[2000];
+	char buf2[2000];
 	if (xref) {
 		for (int ti=0; ti<n_file; ti++) {
 			Tlsh *th = tptr[ti];
@@ -427,7 +491,10 @@ int max_files;
 							if (show_details)
 								printf("%s	[%s]	%s	[%s]	%d\n", fnames[ti].name, th->getHash(), fnames[xi].name, xh->getHash(), tdiff);
 							else
-								printf("%s	%s	%d\n", fnames[ti].name, fnames[xi].name, tdiff);
+								printf("%s	%s	%d\n", convert_special_chars(fnames[ti].name, buf1, sizeof(buf1), 
+							 	                                             "\t", CONVERT_TAB, "\n", CONVERT_NEWLINE, "\r", CONVERT_LINEFEED), 
+							   	                       convert_special_chars(fnames[xi].name, buf2, sizeof(buf2),
+								                                             "\t", CONVERT_TAB, "\n", CONVERT_NEWLINE, "\r", CONVERT_LINEFEED), tdiff);
 						}
 					}
 				}
@@ -441,13 +508,18 @@ int max_files;
 					int tdiff = comp_th->totalDiff(th, xlen);
 					if (tdiff <= threshold) {
 						if (dirname || listname) {
-							printf("%s	%s	%d\n", compare_fname, fnames[ti].name, tdiff);
+							printf("%s	%s	%d\n", convert_special_chars(compare_fname, buf1, sizeof(buf1), 
+						 	                                             "\t", CONVERT_TAB, "\n", CONVERT_NEWLINE, "\r", CONVERT_LINEFEED), 
+							                       convert_special_chars(fnames[ti].name, buf2, sizeof(buf2),
+						 	                                             "\t", CONVERT_TAB, "\n", CONVERT_NEWLINE, "\r", CONVERT_LINEFEED), tdiff);
 						} else {
-							printf("%4d	%s\n", tdiff, fnames[ti].name);
+							printf("%4d	%s\n", tdiff, convert_special_chars(fnames[ti].name, buf1, sizeof(buf1),
+							                                                "\t", CONVERT_TAB, "\n", CONVERT_NEWLINE, "\r", CONVERT_LINEFEED)); 
 						}
 					}
 				} else {
-					printf("%s	%s\n", th->getHash(), fnames[ti].name);
+					printf("%s	%s\n", th->getHash(), convert_special_chars(fnames[ti].name, buf1, sizeof(buf1),
+							                                                "\t", CONVERT_TAB, "\n", CONVERT_NEWLINE, "\r", CONVERT_LINEFEED)); 
 				}
 			}
 		}
@@ -504,7 +576,7 @@ static void usage(const char *fullPathName)
 	printf("  -xlen:              Passed as the len_diff parameter to Tlsh::totalDiff().  If not specified, len_diff will be true, else false.  Determines if the lengths\n");
 	printf("                      of the compared files is to be included in determining the distance.  See tlsh.h for details.\n");
 	printf("  -details:           Results in extra detailed output.\n");
-	printf("  -T threshold_value: Used only during comparisons (-c file|digset or -xref).  Specifies the maximun distance that a comparison must\n"); 
+	printf("  -T threshold_value: Used only during comparisons (-c file|digset or -xref).  Specifies the maximum distance that a comparison must\n"); 
 	printf("                      generate before it is reported. (defaults to %d)\n", DEFAULT_THRESHOLD);
 	printf("\n");
 	printf("Example usages:\n");
